﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Optimization;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  class HypervolumeAnalyzer : MOTFAnalyzer {
    public HypervolumeAnalyzer() {
    }

    protected HypervolumeAnalyzer(HypervolumeAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new HypervolumeAnalyzer(this, cloner);
    }

    protected override void Analyze(Individual[] individuals, double[][] qualities, ResultCollection results) {
      if (qualities == null || qualities.Length < 2) return;
      int objectives = qualities[0].Length;
      double best = TestFunction.BestKnownHypervolume(objectives);

      double diff;

      if (!results.ContainsKey("Hypervolume")) results.Add(new Result("Hypervolume", typeof(DoubleValue)));
      IEnumerable<double[]> front = NonDominatedSelect.selectNonDominatedVectors(qualities, TestFunction.Maximization(objectives), true);
      if (!results.ContainsKey("BestKnownHypervolume")) results.Add(new Result("BestKnownHypervolume", typeof(DoubleValue)));
      else {
        DoubleValue dv = (DoubleValue)(results["BestKnownHypervolume"].Value);
        best = dv.Value;
      }
      if (!results.ContainsKey("Absolute Distance to BestKnownHypervolume")) results.Add(new Result("Absolute Distance to BestKnownHypervolume", typeof(DoubleValue)));

      double hv = Double.NaN;
      try {
        if (objectives == 2) { //Hypervolume analysis only with 2 objectives for now
          hv = Hypervolume.Calculate(front, TestFunction.ReferencePoint(objectives), TestFunction.Maximization(objectives));
        } else if (Array.TrueForAll(TestFunction.Maximization(objectives), x => !x)) {
          hv = FastHV2.Calculate(front, TestFunction.ReferencePoint(objectives));
        }
      }
      catch (ArgumentException) {
        //TODO 
      }

      if (best < 0) {
        best = hv;
      }
      diff = best - hv;
      if (!Double.IsNaN(hv) && (diff < 0 || best < 0)) {
        best = hv;
        diff = 0;
      }

      results["Hypervolume"].Value = new DoubleValue(hv);
      results["Absolute Distance to BestKnownHypervolume"].Value = new DoubleValue(diff);
      results["BestKnownHypervolume"].Value = new DoubleValue(best);

    }
  }
}
